"""Module containing Merkle Tree implementations"""
# pylint: disable=no-value-for-parameter
from typing import List, Dict
from hexbytes import HexBytes
from eth_typing import HexStr
from web3 import Web3


class MerkleTreeKeccak:
    """Class for computing a sorted Merkle Tree using Keccak hashing algorithm"""

    def __init__(self, leaves: List[HexBytes]):
        MerkleTreeKeccak.__check_input(leaves)
        self.tree = MerkleTreeKeccak.__compute_tree(leaves)

    @staticmethod
    def __check_input(leaves: List[HexBytes]):
        if not isinstance(leaves, list):
            raise ValueError("MerkleTreeKeccak: Leaves are not a list")

        if len(leaves) == 0:
            raise ValueError("MerkleTreeKeccak: Leaves are an empty list")

        for leaf in leaves:
            if not isinstance(leaf, HexBytes):
                raise TypeError(f"MerkleTreeKeccak: {leaf!r} not of HexBytes type")
            if len(leaf) != 32:
                raise ValueError(
                    f"MerkleTreeKeccak: {leaf!r} is not a valid 32 byte hash"
                )

    @staticmethod
    def __compute_tree(leaves: List[HexBytes]):
        tree: Dict = {}
        leaves.sort()
        tree[0] = leaves
        node_level = 0
        while len(tree[node_level]) != 1:
            nodes = tree[node_level]
            next_nodes: List[HexBytes] = []
            num_of_nodes = len(nodes)
            is_odd = num_of_nodes % 2 != 0
            if is_odd:
                for i in range(0, num_of_nodes - 1, 2):
                    next_nodes.append(
                        Web3.solidityKeccak(
                            ["bytes32", "bytes32"], [nodes[i], nodes[i + 1]]
                        )
                    )
                next_nodes.append(nodes[-1])
            else:
                for i in range(0, num_of_nodes, 2):
                    next_nodes.append(
                        Web3.solidityKeccak(
                            ["bytes32", "bytes32"], [nodes[i], nodes[i + 1]]
                        )
                    )
            # sort in place
            next_nodes.sort()
            tree[node_level + 1] = next_nodes
            node_level = len(tree.keys()) - 1
        return tree

    @property
    def root_hash(self) -> HexStr:
        """Returns the merkle root hash"""
        return self.tree[len(self.tree.keys()) - 1][0].hex()

    def get_proof(self, leaf: HexBytes) -> List[HexStr]:
        """Function to get the merkle proof for an input leaf"""
        if not isinstance(leaf, HexBytes):
            raise TypeError(f"MerkleTreeKeccak: {leaf!r} not of HexBytes type")
        if len(leaf) != 32:
            raise ValueError(f"MerkleTreeKeccak: {leaf!r} is not a valid 32 byte hash")

        if self.tree[0].index(leaf) < 0:
            print(self.tree[0].index(leaf))
            return []

        proof: List[HexStr] = []
        num_of_nodes: int = len(self.tree.keys()) - 1  # don't need the root hash
        lookup_val: HexBytes = leaf
        for i in range(num_of_nodes):
            loc: int = 0
            for (node, j) in zip(self.tree[i], range(len(self.tree[i]))):
                if node.hex() == lookup_val.hex():
                    loc = j
                    break
            if loc % 2 == 0 and loc != len(self.tree[i]) - 1:
                proof.append(self.tree[i][loc + 1].hex())
                lookup_val = Web3.solidityKeccak(
                    ["bytes32", "bytes32"],
                    [self.tree[i][loc], self.tree[i][loc + 1]],
                )
            elif loc % 2 == 0 and loc == len(self.tree[i]) - 1:
                pass
            else:
                proof.append(self.tree[i][loc - 1].hex())
                lookup_val = Web3.solidityKeccak(
                    ["bytes32", "bytes32"],
                    [self.tree[i][loc - 1], self.tree[i][loc]],
                )
        return proof



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['evm_sc_utils']

package_data = \
{'': ['*']}

install_requires = \
['web3>=5.29.0,<6.0.0']

setup_kwargs = {
    'name': 'evm-sc-utils',
    'version': '0.1.0',
    'description': 'Utilities that are helpful when developing Ethereum Virtual Machine Smart Contracts with Python.',
    'long_description': None,
    'author': 'mpeyfuss',
    'author_email': '82735893+mpeyfuss@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



"""Module containing signer classes."""
# pylint: disable=no-value-for-parameter
from typing import List
from eth_typing import Address
from eth_account import Account
from eth_account.messages import encode_defunct, SignableMessage
from eth_account.datastructures import SignedMessage
from hexbytes import HexBytes


class EIP191Signer:
    """Signer for EIP-191 messages"""

    def __init__(self, private_key: Address = None) -> None:
        if private_key is None:
            self.acc = Account.create()
        else:
            if len(HexBytes(private_key)) == 32:
                self.acc = Account.from_key(private_key)
            else:
                raise ValueError("EIP191Signer: private_key not a 20 byte address")

    @staticmethod
    def __check_inputs(abi_types: List[str], values: List) -> None:
        if not isinstance(abi_types, list) or len(abi_types) == 0:
            raise ValueError(
                "EIP191Signer:sign abi_types cannot be None or an empty list"
            )

        if not isinstance(values, list) or len(values) == 0:
            raise ValueError("EIP191Signer:sign values cannot be None or an empty list")

        if len(abi_types) != len(values):
            raise ValueError(
                "EIP191Signer:sign length mismatch between abi_types and values"
            )

    @staticmethod
    def __pack_data(abi_types: List[str], values: List) -> bytearray:
        data: bytearray = bytearray()
        length: int = 0
        val: int = 0
        for abi_type, value in zip(abi_types, values):
            if "int8" in abi_type.lower():
                length = 1
                val = value
            elif "int16" in abi_type.lower():
                length = 2
                val = value
            elif "int32" in abi_type.lower():
                length = 4
                val = value
            elif "int64" in abi_type.lower():
                length = 8
                val = value
            elif "int128" in abi_type.lower():
                length = 16
                val = value
            elif "int256" in abi_type.lower():
                length = 32
                val = value
            elif "bytes32" in abi_type.lower():
                length = 32
                val = int(value, 16)
            elif "address" in abi_type.lower():
                length = 20
                val = int(value, 16)
            elif "string" in abi_type.lower():
                length = len(value)
                data.extend(value.encode("utf-8"))  # extend in place

            if isinstance(val, int):
                data.extend(bytearray(val.to_bytes(length=length, byteorder="big")))

        return data

    def sign(self, abi_types: List[str], values: List) -> SignedMessage:
        """Function to sign pack and sign the data supplied according to
        the ABI types and EIP-191 version 0x45"""
        EIP191Signer.__check_inputs(abi_types, values)
        msg: SignableMessage = self.get_signable_message(abi_types, values)
        return Account.sign_message(msg, private_key=self.acc.key)

    @staticmethod
    def get_signable_message(abi_types: List[str], values: List) -> SignableMessage:
        """Function to get the signable message by packing the data according to abi types"""
        EIP191Signer.__check_inputs(abi_types, values)
        data: bytearray = EIP191Signer.__pack_data(abi_types, values)
        return encode_defunct(data)

    @property
    def address(self) -> str:
        """Returns the public key of the signer"""
        return self.acc.address
